import cleanPatch from '@thatamplify/core/api/utils/cleanPatch'
import Tenant from '@thatamplify/core/app/Tenant/Tenant'
import DataStoreError from '@thatamplify/core/errors/DataStoreError'
import { useDebounce1000 } from '@thatreact/core-hooks/hooks/useDebounce'
import catchAndThrow from '@thats/utils/errors/catchAndThrow'
import { DataStore } from 'aws-amplify/datastore'
import { useCallback, useEffect, useRef, useState } from 'react'
import { useApplifyTenantId } from '../../controllers/ApplifyController'
import useDataStoreGet from '../../hooks/useDataStoreGet'
import reg from '../reg'

export default useTenantDAO

export interface TenantDAO {
	tenant: Tenant
	updateTenant: (patch: Partial<Tenant>) => void
	tenantReady: boolean
}

export function useTenantDAO(id: string): TenantDAO {
	const tenantId = useApplifyTenantId()
	const { _, isLoading } = useDataStoreGet(reg.Tenant, id)
	const patch = useRef<Partial<Tenant>>(undefined)
	const cachedItem = useRef<Tenant>(undefined)
	const cachedItemVersion = useRef<number>(0)
	const [cacheChanged, setCacheChanged] = useState(new Date())

	useEffect(() => {
		cachedItem.current = undefined
		cachedItemVersion.current = 0
		patch.current = undefined
		setCacheChanged(new Date())
	}, [id])

	const triggerDataStoreUpdate = useDebounce1000(async () => {
		const _patch = patch.current
		patch.current = undefined
		if (!_patch) return
		const latestItem = (
			await DataStore.query(reg.Tenant, (p: any) => p.tenantId.eq(tenantId) && p.id.eq(id))
		)[0]
		cleanPatch(_patch)
		await DataStore.save(reg.Tenant.copyOf(latestItem, u => Object.assign(u, _patch))).catch(
			catchAndThrow('updating Tenant', DataStoreError)
		)
	}, [tenantId, id])

	const updateTenant = useCallback(
		(_patch: Partial<Tenant>) => {
			const alreadyHasAPatch = !!patch.current

			// Update the update patch to list of DataStore updates
			// NOTE: This means that any upcommited patches will be ditched in favour
			// of the new patch
			patch.current = _patch

			// Perform optimistic update
			cachedItem.current = reg.Tenant.copyOf(cachedItem.current, u => {
				Object.assign(u, _patch)
			})
			if (!alreadyHasAPatch) cachedItemVersion.current++

			// Force a refresh
			setCacheChanged(new Date())

			triggerDataStoreUpdate()
		},
		[triggerDataStoreUpdate]
	)

	useEffect(() => {
		if (!!patch.current) return
		if (cachedItemVersion?.current >= _?.['_version']) return
		cachedItem.current = _
		cachedItemVersion.current = _?.['_version'] ?? undefined
		setCacheChanged(new Date())
	}, [_])

	return {
		tenant: cachedItem.current,
		updateTenant,
		tenantReady: !isLoading && !!cachedItem.current,
	}
}
