import cleanPatch from '@thatamplify/core/api/utils/cleanPatch'
import TenantUser from '@thatamplify/core/app/TenantUser/TenantUser'
import DataStoreError from '@thatamplify/core/errors/DataStoreError'
import { useDebounce1000 } from '@thatreact/core-hooks/hooks/useDebounce'
import catchAndThrow from '@thats/utils/errors/catchAndThrow'
import { DataStore } from 'aws-amplify/datastore'
import { useCallback, useEffect, useRef, useState } from 'react'
import { useApplifyTenantId } from '../../controllers/ApplifyController'
import useDataStoreGet from '../../hooks/useDataStoreGet'
import reg from '../reg'

export default useTenantUserDAO

export interface TenantUserDAO {
	tenantUser: TenantUser
	updateTenantUser: (patch: Partial<TenantUser>) => void
	tenantUserReady: boolean
}

export function useTenantUserDAO(id: string): TenantUserDAO {
	const tenantId = useApplifyTenantId()
	const { _, isLoading } = useDataStoreGet(reg.TenantUser, id)
	const patch = useRef<Partial<TenantUser>>(undefined)
	const cachedItem = useRef<TenantUser>(undefined)
	const cachedItemVersion = useRef<number>(0)
	const [cacheChanged, setCacheChanged] = useState(new Date())

	useEffect(() => {
		cachedItem.current = undefined
		cachedItemVersion.current = 0
		patch.current = undefined
		setCacheChanged(new Date())
	}, [id])

	const triggerDataStoreUpdate = useDebounce1000(async () => {
		const _patch = patch.current
		patch.current = undefined
		if (!_patch) return
		const latestItem = (
			await DataStore.query(
				reg.TenantUser,
				(p: any) => p.tenantId.eq(tenantId) && p.id.eq(id)
			)
		)[0]
		cleanPatch(_patch)
		await DataStore.save(
			reg.TenantUser.copyOf(latestItem, u => Object.assign(u, _patch))
		).catch(catchAndThrow('updating TenantUser', DataStoreError))
	}, [tenantId, id])

	const updateTenantUser = useCallback(
		(_patch: Partial<TenantUser>) => {
			const alreadyHasAPatch = !!patch.current

			// Update the update patch to list of DataStore updates
			// NOTE: This means that any upcommited patches will be ditched in favour
			// of the new patch
			patch.current = _patch

			// Perform optimistic update
			cachedItem.current = reg.TenantUser.copyOf(cachedItem.current, u => {
				Object.assign(u, _patch)
			})
			if (!alreadyHasAPatch) cachedItemVersion.current++

			// Force a refresh
			setCacheChanged(new Date())

			triggerDataStoreUpdate()
		},
		[triggerDataStoreUpdate]
	)

	useEffect(() => {
		if (!!patch.current) return
		if (cachedItemVersion?.current >= _?.['_version']) return
		cachedItem.current = _
		cachedItemVersion.current = _?.['_version'] ?? undefined
		setCacheChanged(new Date())
	}, [_])

	return {
		tenantUser: cachedItem.current,
		updateTenantUser,
		tenantUserReady: !isLoading && !!cachedItem.current,
	}
}
