import cleanPatch from '@thatamplify/core/api/utils/cleanPatch'
import TenantUser from '@thatamplify/core/app/TenantUser/TenantUser'
import TenantUserPS from '@thatamplify/core/app/_propsets/TenantUserPS'
import RecordType from '@thatamplify/core/constants/RecordType'
import DataStoreError from '@thatamplify/core/errors/DataStoreError'
import { useDebounce1000 } from '@thatreact/core-hooks/hooks/useDebounce'
import { swap } from '@thats/utils/array/swap'
import catchAndThrow from '@thats/utils/errors/catchAndThrow'
import { DataStore } from 'aws-amplify/datastore'
import { useCallback, useEffect, useRef, useState } from 'react'
import { useApplifyTenantId } from '../../controllers/ApplifyController'
import { useDataStoreList } from '../../hooks/useDataStoreList'
import reg from '../reg'

export default useTenantUsersDAO

export type TenantUserPatch = { name: string; p: Partial<TenantUserPS<RecordType.FLAT>> }

export interface TenantUsersDAO {
	tenantUsers: TenantUser[]

	updateTenantUser: (id: string, patch: TenantUserPatch) => void
	updateTenantUserPS: (id: string, patch: Partial<TenantUserPS<RecordType.FLAT>>) => void
	deleteTenantUser: (itemOrItems: TenantUser | TenantUser[]) => void
	tenantUsersReady: boolean
}

export function useTenantUsersDAO(): TenantUsersDAO {
	const { _, isLoading } = useDataStoreList(reg.TenantUser)

	const tenantId = useApplifyTenantId()
	const itemCache = useRef<TenantUser[]>([])

	const [cacheChanged, setCacheChanged] = useState(new Date())

	const noOfPatches = useRef(0)
	const patches = useRef<Record<string, TenantUserPatch[]>>({})
	const triggerDataStoreUpdate = useDebounce1000(async () => {
		const p = patches.current
		for (const id of Object.keys(p)) {
			const patchesBeingProcessed = [...p[id]]
			for (const patch of p[id]) {
				const latestItem = (
					await DataStore.query(
						reg.TenantUser,
						(p: any) => p.tenantId.eq(tenantId) && p.id.eq(id)
					)
				)[0]
				cleanPatch(patch)
				await DataStore.save(
					reg.TenantUser.copyOf(latestItem, u => Object.assign(u, patch))
				).catch(catchAndThrow('updating TenantUser', DataStoreError))
				noOfPatches.current--
			}
			p[id] = p[id].filter(pp => patchesBeingProcessed.indexOf(pp) < 0)
		}
	}, [tenantId])

	const updateTenantUser = useCallback(
		(id: string, patch: TenantUserPatch) => {
			// Perform optimistic update
			const ic = itemCache.current
			const idx = ic.findIndex(p => p.id === id)
			const original = ic[idx]
			const patched = reg.TenantUser.copyOf(original, u => {
				Object.assign(u, patch)
				// eslint-disable-next-line @typescript-eslint/ban-ts-comment
				// @ts-ignore
				u['_version']++
			})
			itemCache.current = swap(ic, patched, x => x === original)

			// Force a refresh
			setCacheChanged(new Date())

			// Add Patch to list of DataStore updates
			const p = patches.current
			if (!p[id]) p[id] = [] as TenantUserPatch[]
			p[id].push(patch)
			noOfPatches.current++
			triggerDataStoreUpdate()
		},
		[triggerDataStoreUpdate]
	)

	const updateTenantUserPS = useCallback(
		(id: string, patch: Partial<TenantUserPS<RecordType.FLAT>>) => {
			const ic = itemCache.current
			const idx = ic.findIndex(p => p.id === id)
			const original = ic[idx]

			updateTenantUser(id, {
				...original,
				p: { ...original.p, ...patch },
			})
		},
		[updateTenantUser]
	)

	const itemsToDelete = useRef<TenantUser[]>([])
	const triggerDataStoreDelete = useDebounce1000(async () => {
		const items = itemsToDelete.current
		for (const item of items) {
			const latestItem = (
				await DataStore.query(
					reg.TenantUser,
					(p: any) => p.tenantId.eq(tenantId) && p.id.eq(item.id)
				)
			)[0]
			await DataStore.delete(latestItem).catch(
				catchAndThrow('deleting TenantUser', DataStoreError)
			)
			itemsToDelete.current = itemsToDelete.current.filter(i => i.id !== item.id)
		}
	}, [tenantId])

	const deleteTenantUser = useCallback(
		(itemOrItems: TenantUser | TenantUser[]) => {
			const items = Array.isArray(itemOrItems) ? itemOrItems : [itemOrItems]
			for (const item of items) {
				// Perform optimistic delete
				itemCache.current = itemCache.current.filter(i => i.id !== item.id)

				// Force a refresh
				setCacheChanged(new Date())

				// Add Item to list of Items to delete
				itemsToDelete.current.push(item)
			}
			triggerDataStoreDelete()
		},
		[triggerDataStoreDelete]
	)

	useEffect(() => {
		if (noOfPatches.current > 0) return
		if (itemsToDelete.current.length > 0) return

		// Only update if:
		// 1) the number of items returned has changed
		// OR
		// 2) an item with a LARGER version has been sent
		if (
			_?.length !== itemCache?.current.length ||
			_.some(item => {
				const cachedItem = itemCache.current.find(i => i.id === item.id)
				if (!cachedItem) return true
				const cv = cachedItem['_version']
				const iv = item['_version']
				if (cv === undefined && iv === 1) return false
				return iv > cv
			})
		) {
			itemCache.current = _
			setCacheChanged(new Date())
		}
	}, [_])

	return {
		tenantUsers: itemCache.current,

		updateTenantUser,
		updateTenantUserPS,
		deleteTenantUser,
		tenantUsersReady: !isLoading,
	}
}
