import cleanPatch from '@thatamplify/core/api/utils/cleanPatch'
import TenantUserInvite from '@thatamplify/core/app/TenantUserInvite/TenantUserInvite'
import TenantUserInvitePS from '@thatamplify/core/app/_propsets/TenantUserInvitePS'
import RecordType from '@thatamplify/core/constants/RecordType'
import DataStoreError from '@thatamplify/core/errors/DataStoreError'
import { useDebounce1000 } from '@thatreact/core-hooks/hooks/useDebounce'
import { swap } from '@thats/utils/array/swap'
import catchAndThrow from '@thats/utils/errors/catchAndThrow'
import { DataStore } from 'aws-amplify/datastore'
import { useCallback, useEffect, useRef, useState } from 'react'
import { useApplifyTenantId } from '../../controllers/ApplifyController'
import { useDataStoreList } from '../../hooks/useDataStoreList'
import reg from '../reg'

export default useTenantUserInvitesDAO

export type TenantUserInvitePatch = {
	name: string
	p: Partial<TenantUserInvitePS<RecordType.FLAT>>
}

export interface TenantUserInvitesDAO {
	tenantUserInvites: TenantUserInvite[]

	updateTenantUserInvite: (id: string, patch: TenantUserInvitePatch) => void
	updateTenantUserInvitePS: (
		id: string,
		patch: Partial<TenantUserInvitePS<RecordType.FLAT>>
	) => void
	deleteTenantUserInvite: (itemOrItems: TenantUserInvite | TenantUserInvite[]) => void
	tenantUserInvitesReady: boolean
}

export function useTenantUserInvitesDAO(): TenantUserInvitesDAO {
	const { _, isLoading } = useDataStoreList(reg.TenantUserInvite)

	const tenantId = useApplifyTenantId()
	const itemCache = useRef<TenantUserInvite[]>([])

	const [cacheChanged, setCacheChanged] = useState(new Date())

	const noOfPatches = useRef(0)
	const patches = useRef<Record<string, TenantUserInvitePatch[]>>({})
	const triggerDataStoreUpdate = useDebounce1000(async () => {
		const p = patches.current
		for (const id of Object.keys(p)) {
			const patchesBeingProcessed = [...p[id]]
			for (const patch of p[id]) {
				const latestItem = (
					await DataStore.query(
						reg.TenantUserInvite,
						(p: any) => p.tenantId.eq(tenantId) && p.id.eq(id)
					)
				)[0]
				cleanPatch(patch)
				await DataStore.save(
					reg.TenantUserInvite.copyOf(latestItem, u => Object.assign(u, patch))
				).catch(catchAndThrow('updating TenantUserInvite', DataStoreError))
				noOfPatches.current--
			}
			p[id] = p[id].filter(pp => patchesBeingProcessed.indexOf(pp) < 0)
		}
	}, [tenantId])

	const updateTenantUserInvite = useCallback(
		(id: string, patch: TenantUserInvitePatch) => {
			// Perform optimistic update
			const ic = itemCache.current
			const idx = ic.findIndex(p => p.id === id)
			const original = ic[idx]
			const patched = reg.TenantUserInvite.copyOf(original, u => {
				Object.assign(u, patch)
				// eslint-disable-next-line @typescript-eslint/ban-ts-comment
				// @ts-ignore
				u['_version']++
			})
			itemCache.current = swap(ic, patched, x => x === original)

			// Force a refresh
			setCacheChanged(new Date())

			// Add Patch to list of DataStore updates
			const p = patches.current
			if (!p[id]) p[id] = [] as TenantUserInvitePatch[]
			p[id].push(patch)
			noOfPatches.current++
			triggerDataStoreUpdate()
		},
		[triggerDataStoreUpdate]
	)

	const updateTenantUserInvitePS = useCallback(
		(id: string, patch: Partial<TenantUserInvitePS<RecordType.FLAT>>) => {
			const ic = itemCache.current
			const idx = ic.findIndex(p => p.id === id)
			const original = ic[idx]

			updateTenantUserInvite(id, {
				...original,
				p: { ...original.p, ...patch },
			})
		},
		[updateTenantUserInvite]
	)

	const itemsToDelete = useRef<TenantUserInvite[]>([])
	const triggerDataStoreDelete = useDebounce1000(async () => {
		const items = itemsToDelete.current
		for (const item of items) {
			const latestItem = (
				await DataStore.query(
					reg.TenantUserInvite,
					(p: any) => p.tenantId.eq(tenantId) && p.id.eq(item.id)
				)
			)[0]
			await DataStore.delete(latestItem).catch(
				catchAndThrow('deleting TenantUserInvite', DataStoreError)
			)
			itemsToDelete.current = itemsToDelete.current.filter(i => i.id !== item.id)
		}
	}, [tenantId])

	const deleteTenantUserInvite = useCallback(
		(itemOrItems: TenantUserInvite | TenantUserInvite[]) => {
			const items = Array.isArray(itemOrItems) ? itemOrItems : [itemOrItems]
			for (const item of items) {
				// Perform optimistic delete
				itemCache.current = itemCache.current.filter(i => i.id !== item.id)

				// Force a refresh
				setCacheChanged(new Date())

				// Add Item to list of Items to delete
				itemsToDelete.current.push(item)
			}
			triggerDataStoreDelete()
		},
		[triggerDataStoreDelete]
	)

	useEffect(() => {
		if (noOfPatches.current > 0) return
		if (itemsToDelete.current.length > 0) return

		// Only update if:
		// 1) the number of items returned has changed
		// OR
		// 2) an item with a LARGER version has been sent
		if (
			_?.length !== itemCache?.current.length ||
			_.some(item => {
				const cachedItem = itemCache.current.find(i => i.id === item.id)
				if (!cachedItem) return true
				const cv = cachedItem['_version']
				const iv = item['_version']
				if (cv === undefined && iv === 1) return false
				return iv > cv
			})
		) {
			itemCache.current = _
			setCacheChanged(new Date())
		}
	}, [_])

	return {
		tenantUserInvites: itemCache.current,

		updateTenantUserInvite,
		updateTenantUserInvitePS,
		deleteTenantUserInvite,
		tenantUserInvitesReady: !isLoading,
	}
}
