import IDMap from '@thats/utils/map/IDMap'
import TenantedTINP from '@thats/utils/types/TenantedTINP'
import { ModelInit, PersistentModelConstructor as PMC } from 'aws-amplify/datastore'
import { useCallback, useEffect, useMemo, useRef, useState } from 'react'
import * as uuid from 'uuid'
import { useApplifyTenantId } from '../controllers/ApplifyController'
import gqlMutation from '../utils/gqlMutation'
import gqlSubscribe from '../utils/gqlSubscribe'
import gqlify from './gqlify'

export const cache: { [key: string]: IDMap<any> } = {}
export const subscribers: (() => void)[] = []

export type DataPatch = {
	p?: any
}

export type DataCacheGQL = {
	create: any
	update: any
	delete: any
	onCreate: any
	onDelete: any
	onUpdate: any
}

export function useDataCacheByProjectId<
	T extends TenantedTINP = TenantedTINP,
	PATCH extends DataPatch = DataPatch
>(PM: PMC<T>, gql: DataCacheGQL, projectId: string) {
	const tenantId = useApplifyTenantId()
	const [cacheChanged, setCacheChanged] = useState(new Date())
	const cacheKey = useRef(`${tenantId}:${projectId}:${PM.name}`)
	const subscriptions = useRef<any>({})

	const handleChange = useCallback(() => setCacheChanged(new Date()), [setCacheChanged])
	const triggerChange = useCallback(() => subscribers.forEach(cb => cb()), [])

	useEffect(() => {
		subscribers.push(handleChange)
		return () => {
			const index = subscribers.indexOf(handleChange)
			if (index > -1) subscribers.splice(index, 1)
		}
	}, [handleChange])

	const update = useCallback(
		(id: string, patch: PATCH) => {
			const item = cache[cacheKey.current].get(id)
			const newItem = PM.copyOf(item, u =>
				Object.assign(u, {
					...u,
					...patch,
					p: {
						...(u['p'] ?? {}),
						...(patch['p'] ?? {}),
					},
				})
			)
			void gqlMutation(gql.update, gqlify(newItem))
			cache[cacheKey.current].set(newItem.id, newItem)
			triggerChange()
		},
		[PM, gql.update, triggerChange]
	)

	const create = useCallback(
		(init: ModelInit<T>) => {
			const id = uuid.v4()
			const newItem = new PM({ tenantId, id, ...init } as any)
			cache[cacheKey.current].set(newItem.id, newItem)
			void gqlMutation(gql.create, gqlify(newItem))
			triggerChange()
			return newItem
		},
		[PM, gql.create, tenantId, triggerChange]
	)

	const deleteByIds = useCallback(
		(ids: string[]) => {
			ids.forEach(id => {
				void gqlMutation(gql.delete, { input: { id } })
				cache[cacheKey.current].delete(id)
			})
			triggerChange()
		},
		[gql.delete, triggerChange]
	)

	const set = useCallback(
		(item: T) => {
			if (!item) return
			if (!cache[cacheKey.current]) cache[cacheKey.current] = new IDMap()
			cache[cacheKey.current].set(item.id, item)
			triggerChange()
		},
		[triggerChange]
	)

	const setAll = useCallback(
		(items: T[]) => {
			if (!items) return
			if (!cache[cacheKey.current]) cache[cacheKey.current] = new IDMap()
			items.forEach(item => cache[cacheKey.current].set(item.id, item))
			triggerChange()
		},
		[triggerChange]
	)

	const get = useCallback(
		(id: string) => (cache[cacheKey.current]?.get(id) ?? undefined) as T,
		[]
	)

	const getAll = useCallback(
		() => (cache[cacheKey.current]?.valuesArr() ?? []) as T[],
		[cacheKey]
	)

	useEffect(() => {
		const s = subscriptions.current
		const k = `${tenantId}:${projectId}:${PM.name}`
		// Subscribe to create
		s[`${k}:create`] = gqlSubscribe(gql.onCreate, { tenantId, projectId }, async v => {
			const newItem = PM['fromJSON'](v.data[`onCreate${PM.name}ByProjectId`])
			cache[k].set(newItem.id, newItem)
			triggerChange()
		})

		// Subscribe to delete
		s[`${k}:delete`] = gqlSubscribe(gql.onDelete, { tenantId, projectId }, async v => {
			const id = v.data[`onDelete${PM.name}ByProjectId`].id
			cache[k].delete(id)
			triggerChange()
		})

		// Subscribe to update
		s[`${k}:update`] = gqlSubscribe(gql.onUpdate, { tenantId, projectId }, async v => {
			const updatedItem = PM['fromJSON'](v.data[`onUpdate${PM.name}ByProjectId`])
			cache[k].set(updatedItem.id, updatedItem)
			triggerChange()
		})
		return () => {
			Object.keys(s).forEach(key => {
				try {
					s[key]?.['unsubscribe']?.()
					delete s[key]
				} catch {}
			})
		}
	}, [PM, gql, projectId, tenantId, triggerChange])

	useEffect(() => {
		cacheKey.current = `${tenantId}:${projectId}:${PM.name}`
	}, [PM.name, projectId, tenantId])

	return useMemo(
		() => ({
			cacheChanged,
			get,
			getAll,
			set,
			setAll,
			update,
			create,
			deleteByIds,
		}),
		[cacheChanged, create, deleteByIds, get, getAll, set, setAll, update]
	)
}

export default useDataCacheByProjectId
