import useDebounce from '@thatreact/core-hooks/hooks/useDebounce'
import Logger from '@thats/utils/log/Logger'
import { DataStore, PersistentModel, PersistentModelConstructor } from 'aws-amplify/datastore'
import { useEffect, useRef, useState } from 'react'
import {
	useApplifyActions,
	useApplifyCtxChanged,
	useApplifyTenantId,
	useApplifyWaitForSynced,
} from '../controllers/ApplifyController'

export default useDataStoreGet

export type UseDataStoreGetResult<PM extends PersistentModel> = {
	_: PM
	isLoading: boolean
	isSynced: boolean
}

export function useDataStoreGet<PM extends PersistentModel>(
	PM: PersistentModelConstructor<PM>,
	id: string,
	o: {
		skip?: boolean
		log?: boolean
	} = {}
): UseDataStoreGetResult<PM> {
	const { skip = false, log = false } = o
	const subscription = useRef<any>()
	const waitForSynced = useApplifyWaitForSynced()
	const tenantId = useApplifyTenantId()
	const ctxChanged = useApplifyCtxChanged()
	const { doDataStoreSubscribe, doDataStoreUnsubscribe } = useApplifyActions()
	const [result, setResult] = useState<UseDataStoreGetResult<PM>>({
		_: undefined,
		isLoading: !skip,
		isSynced: false,
	})

	const doSetResult = useDebounce(newResult => {
		setResult(newResult)
	}, [])

	useEffect(() => {
		if (skip || !id) return
		if (subscription.current) doDataStoreUnsubscribe(subscription.current)

		subscription.current = doDataStoreSubscribe(
			DataStore.observeQuery(
				PM,
				(p: any) => p.tenantId.eq(tenantId) && p.id.eq(id)
			).subscribe(snapshot => {
				const { items, isSynced } = snapshot
				if (log) Logger.info('Snapshot Received', { items, isSynced })
				if (!waitForSynced || isSynced)
					doSetResult({
						_: items?.[0] ?? undefined,
						isLoading: false,
						isSynced,
					})
			})
		)
		return () => {
			if (subscription.current) doDataStoreUnsubscribe(subscription.current)
		}
	}, [
		skip,
		PM,
		tenantId,
		id,
		log,
		waitForSynced,
		doSetResult,
		ctxChanged,
		doDataStoreUnsubscribe,
		doDataStoreSubscribe,
	])

	useEffect(() => {
		doSetResult({
			_: undefined,
			isLoading: !skip,
			isSynced: false,
		})
	}, [ctxChanged, doSetResult, skip])

	return result
}
